"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from .metrics import (
    score,
    ndcg,
    normalized_discounted_cumulative_gain,
    precision_at_k,
    coverage_at_k,
    cak,
    pak,
)

from .utils import split, he_sampling, leave_one_out



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from functools import partial

from .pop_rank import PopRankModel
from .mf import MatrixFactorizationModel

AlternatingLeastSquaresModel = partial(MatrixFactorizationModel, method="als")
BayesianPersonalizedRankingModel = partial(MatrixFactorizationModel, method="bpr")



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from typing import Optional
import numpy as np
from numpy import ndarray

from xanthus.datasets import Dataset
from xanthus.models import base


class PopRankModel(base.RecommenderModel):
    """
    A Popularity Ranking recommender model. This class implements some simple
    popularity-based recommendation approaches. By default, it will simply return
    the most popular items that meet a given criteria (e.g. in a given list). It
    also implements a type of weighted shuffle to inject lower-ranked items higher
    in the popularity rankings.

    Parameters
    ----------
    alpha: float
        The 'strength' of weighting towards popular terms: '0' would correspond to
        a uniform shuffle. Only used when sampling="weighted".
        Implemented as in [1].
    limit: int
        The maximum number of items to select for ordering. Increasing this will
        impact performance for large sets of items.
    sampling: str
        The sampling method to use when generating recommendations. This is used to
        shuffle the popularity-ranked items (i.e. the popularity ranked items will be
        reordered according to this sampling method).
            * uniform - randomly shuffle top ranked items.
            * weighted - shuffle items, but tend to keep top ranked items higher in the
                         recommender's outputs.
            * None - well, it seems you don't want to shuffle the items at all! This
                     is 'pure' PopRank.

    References
    ----------
    [1] https://godatadriven.com/blog/elitist-shuffle-for-recommendation-systems/

    """

    def __init__(
        self, alpha: float = 1.0, limit: int = 100, sampling: Optional[str] = "uniform"
    ) -> None:
        """Time to initialise a PopRankModel."""

        self._sampling = sampling
        self._ranked = None
        self._alpha = alpha
        self._limit = limit

    def fit(self, dataset: Dataset) -> "PopRankModel":
        """
        Fit the model to a provided Dataset.

        Parameters
        ----------
        dataset: Dataset
            An input dataset.

        Returns
        -------
        output: PopRankModel
            Returns itself. How fun.

        See Also
        --------
        xanthus.datasets.Dataset

        """

        init_weights = np.asarray(dataset.interactions.sum(axis=0)).flatten()
        ranked = init_weights.argsort()[::-1]
        self._ranked = ranked
        return self

    def predict(
        self,
        dataset: Optional[Dataset] = None,
        users: ndarray = None,
        items: ndarray = None,
        excluded: Optional[ndarray] = None,
        n: int = 3,
    ) -> ndarray:
        """
        Generate predictions (recommendations) from the model. For each provided user,
        the output will be a set of items ranked in order of predicted preference.

        Parameters
        ----------
        dataset: Dataset, optional
            The dataset for which you wish to generate recommendations. If this is
            provided, this object's `all_items` and `users` will be used for the
            purpose of generating recommendations.
        users: list, optional
            An optional array of users for whom you wish to generate recommendations.
        items: list, optional
            An optional array of items you wish to be used in recommendations. This
            may be a subset of items for the purposes of ranking specific subsets of
            items, for example. Maybe you want to see which Star Wars movies are the
            most loved according to the model, for example.
        n: int
            The number of recommendations to be generated per user.
        excluded: list, optional
            An optional array of items to exclude from recommendations.

        Returns
        -------
        output: list
            A list, where each element corresponds to a list of recommendations.

        """

        if excluded is not None:
            valid = np.isin(excluded, self._ranked)
            items = self._ranked[valid]

            if len(items) < n:
                raise ValueError(
                    f"Insufficient valid items available ({len(items)}, required {n})"
                    "after filtering excluded items."
                )

        elif items is not None:
            items = self._ranked[np.isin(items, self._ranked)]

        else:
            items = self._ranked

        # truncate items for filter/shuffle speed.
        items = items[: self._limit]

        if self._sampling == "weighted":
            weights = np.linspace(1, 0, num=len(items), endpoint=False)
            weights = np.power(weights, self._alpha)
            weights /= np.linalg.norm(weights, ord=1)
        else:
            weights = np.ones(len(items), dtype=np.float32)
            weights /= weights.sum()

        recommended = np.zeros((len(users), n), dtype=np.int32)
        for i, user in enumerate(users):
            if self._sampling is None:
                recommended = items[:n]
            else:
                recommended[i] += np.random.choice(
                    items, size=n, replace=False, p=weights
                )

        return recommended



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from typing import Optional, Any, Tuple

from pandas import DataFrame

from .datasets import Dataset, DatasetEncoder, utils
from .evaluate import leave_one_out, split


def create_datasets(
    df: DataFrame,
    user_df: Optional[DataFrame] = None,
    item_df: Optional[DataFrame] = None,
    policy: str = "leave_one_out",
    **kwargs: Optional[Any]
) -> Tuple[Dataset, Dataset]:
    """
    Utility function for creating train and test datasets.

    Parameters
    ----------
    df: DataFrame
        The input dataset. This must have the fields 'user', 'item' and may optionally
        include the field 'rating'.
    item_df: DataFrame, optional
        Optionally provide a DataFrame containing item metadata. This frame should
        include the fields 'item' and 'tag', where 'tag' corresponds to some metadata
        value associated with that item (e.g. genre, director, language in a movie
        use-case).
    user_df: DataFrame, optional
        Optionally provide a DataFrame containing user metadata. This frame should
        include the fields 'user' and 'tag', where 'tag' corresponds to some metadata
        value associated with that user (e.g. age group, location, language in a movie
        use-case).
    policy: str
        The 'split' policy to use when creating 'train' and 'test' sets. By default,
        'leave_one_out' is used. Otherwise, a custom split is used.
    kwargs: optional
        Additional arguments to be passed to a custom 'split' method.

    Returns
    -------
    output: tuple
        A pair where the first element corresponds to a configured 'train' dataset, and
        the second corresponds to a configured 'test' dataset.

    See Also
    --------
    xanthus.evaluate.split

    """

    # split the data into train and test sets
    if policy == "leave_one_out":
        train, test = leave_one_out(df)
    else:
        train, test = split(df, **kwargs)

    # extract item tags (if they're available)
    if item_df is not None:
        item_tags = item_df["tag"]
        items = df["item"].append(item_df["item"])
    else:
        item_tags = None
        items = df["item"]

    # extract user tags (if they're available)
    if user_df is not None:
        user_tags = user_df["tag"]
        users = df["user"].append(user_df["user"])
    else:
        user_tags = None
        users = df["user"]

    # fit the encoder.
    encoder = DatasetEncoder()
    encoder.fit(users=users, items=items, user_tags=user_tags, item_tags=item_tags)

    # build the train dataset.
    train = Dataset.from_df(
        train, encoder=encoder, normalize=utils.as_implicit, item_meta=item_df
    )

    # build the test dataset.
    test = Dataset.from_df(
        test, encoder=encoder, normalize=utils.as_implicit, item_meta=item_df
    )

    # and that's it, you're done!
    return train, test



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

import warnings
from typing import Tuple, List

from pandas import DataFrame, concat

from numpy import (
    in1d,
    concatenate,
    ndarray,
    unique,
    c_,
)
from numpy.random import choice

from ..datasets import Dataset, groupby


def _ignore(df: DataFrame, elements: ndarray, key: str, frac: float) -> DataFrame:
    """
    Remove a random fraction 'frac' of 'elements' with 'key' from a dataframe 'df'.
    """

    ignored_items_size = int(len(elements) * frac)
    ignored_items = choice(elements, size=ignored_items_size, replace=False)
    df = df[~df[key].isin(ignored_items)]
    return df


def leave_one_out(
    df,
    shuffle=True,
    frac_ignored_users: float = 0.0,
    frac_ignored_items: float = 0.0,
    deduplicate: bool = True,
) -> Tuple[DataFrame, ...]:
    """
    Wraps the 'split' function to produce leave-one-out evaluation protocol.
    """

    return split(
        df,
        shuffle=shuffle,
        frac_ignored_users=frac_ignored_users,
        frac_ignored_items=frac_ignored_items,
        deduplicate=deduplicate,
    )


def split(
    df: DataFrame,
    frac_train: float = 0.0,
    shuffle: bool = True,
    n_test: int = 1,
    min_records: int = 1,
    frac_ignored_users: float = 0.0,
    frac_ignored_items: float = 0.0,
    deduplicate: bool = True,
) -> Tuple[DataFrame, ...]:
    """
    A recommendation-system focused train-test split utility function.

    This function was inspired by the Azure ML Studio 'recommender split' utility [1].

    Using default parameters, this function implements the 'leave-one-out' sampling
    approach, whereby a single record from each user is withheld for evaluation
    purposes, and all others included in the training set.

    Parameters
    ----------
    df: DataFrame
        An input dateframe. This frame should represent some set of interactions between
        some 'user' set, and an 'item' set. The fields 'user' and 'item' _must_ be
        present in the frame. As in any CF problem setup, these interactions could be
        page views, purchases, email opens etc.
    frac_train: float
        The fraction of users in the dataset to be used exclusively in the training set.
        None of these users will appear in the test set. The remaining fraction of users
        will be considered 'test candidates', and will be sampled to create the output
        test set, and users in this 'test candidate' set may appear in _both_ the train
        and test set. By default, all users will appear in the test set (frac_train=0).
    shuffle: bool
        Indicate whether the dataset should be shuffled.
    n_test: int
        The number of test samples (from the test candidates) to be sampled. For
        example, if you set `frac_train` to 0.75, 25% of all interactions will be held
        out as test candidates. The function will then sample `n_test` interactions from
        this set. These samples will be come the test set. All remaining candidate test
        samples are appended to the training set. Note that if this is set too high,
        you may have very few 'eligible' users in your candidate set to sample. If there
        are no eligible users, this will raise an error.
    min_records: int
        The minimum number of interactions a user must have had in order to appear in
        the output training data.
    frac_ignored_users: float
        An approximate fraction of users to be ignored. This can be useful to reduce the
        size of large datasets for training purposes.
    frac_ignored_items: float
        An approximate fraction of items to be ignored. This can be useful to reduce the
        size of large datasets for training purposes.
    deduplicate: bool
        Indicate whether to deduplicate interactions (i.e. interactions with the same
        'user' and 'item'). Default is 'True'.

    Returns
    -------
    output: (DataFrame, DataFrame)
        Your specified 'train' and 'test' dataframes.

    References
    ----------
    [1] Inspired by https://docs.microsoft.com/en-us/azure/machine-learning/studio-module-reference/split-data-using-recommender-split#:~:text=The%20Recommender%20Split%20option%20is,user%2Ditem%2Drating%20triples.

    Notes
    -----
    * Benchmarks:
      (0 on 1m records, 500k users, 10k items - ~5s
      (1 on 10m records, 5m users, 10k items - ~60s
      (2 on 20m records, 10m users, 10k items - ~150s
    * Scales roughly linearly in the number of records.

    """

    if shuffle:
        df = df.sample(frac=1.0)

    columns = df.columns

    df["_user"] = df["user"].astype("category").cat.codes
    df["_item"] = df["item"].astype("category").cat.codes

    if deduplicate:
        df = df.drop_duplicates(["_user", "_item"])

    unique_items = df["_item"].unique()
    unique_users = df["_user"].unique()

    if frac_ignored_items > 0:
        df = _ignore(df, unique_items, "_item", frac_ignored_items)

    if frac_ignored_users > 0:
        df = _ignore(df, unique_users, "_user", frac_ignored_users)

    if min_records > 1:
        df = df[df.groupby("_user")["_user"].transform("size") >= min_records]

    # build candidate test frame.
    training_users_size = int(len(unique_users) * frac_train)
    training_users = choice(unique_users, size=training_users_size, replace=False)
    candidate_test_users = unique_users[~in1d(unique_users, training_users)]

    candidate_test_df = df[df["_user"].isin(candidate_test_users)]

    grouped = candidate_test_df.groupby("_user").size() > n_test
    training_users = concatenate(
        [training_users, grouped.index.values[~grouped.values]]
    )
    eligible = grouped.index.values[grouped.values]

    if len(eligible) == 0:
        raise ValueError(
            f"Could not find users in the test sample with more than {n_test} records."
            "Make sure your "
        )

    if len(eligible) < int(len(candidate_test_users) * 0.75):
        warnings.warn(
            "The total number of users that can be sampled from for your test "
            "set is less than 75% of the available users. Consider reducing your"
            f"chosen 'n_test' (it is set to {n_test})."
        )

    # filter the candidate test frame.
    candidate_test_df = candidate_test_df[candidate_test_df["_user"].isin(eligible)]
    excluded_test_df = candidate_test_df[~candidate_test_df["_user"].isin(eligible)]

    # create the training set.
    training_users = concatenate([training_users, excluded_test_df["_user"].unique()])

    train_df = df[df["_user"].isin(training_users)]

    # sample from the candidate test frame.
    test_df_grouped = candidate_test_df.groupby("_user", group_keys=False)
    candidate_test_df["_group"] = test_df_grouped.cumcount()

    test_df = candidate_test_df[candidate_test_df["_group"] < n_test]
    aux_train_df = candidate_test_df[candidate_test_df["_group"] >= n_test]

    # Add interactions that are not in the sampled test frame back to the training set.
    train_df = concat([train_df, aux_train_df])

    return train_df[columns], test_df[columns]


def he_sampling(
    a: Dataset, b: Dataset, n_samples: int = 100
) -> Tuple[ndarray, ndarray]:
    """
    Sample a dataset 'a' with 'n' negative samples given interactions in dataset 'a'
    and 'b'.

    Practically, this function allows you to generate evaluation data as described in
    the work of He et al. [1]. The evaluation procedure assumes that the input datasets
    'a' and 'b' have been generated with a leave 'n' out policy, such that dataset 'b'
    corresponds to the 'training' dataset (i.e. dataset with 'left out' samples removed)
    and 'a' corresponds to the 'test' dataset with 'n' for each user with
    n_interactions > n. For each user in 'a', the function will return that user's 'n'
    left-out interactions, plus 'n_samples' negative samples (items the user has not
    interacted with in both the 'train' and 'test' datasets).

    Parameters
    ----------
    a: Dataset
        The 'test' dataset (the dataset you wish to use for evaluation).
    b: Dataset
        The 'train' dataset (the dataset you wish to include for purposes of sampling
        items the user has not interacted with -- negative samples).
    n_samples: int
        The total number of negative samples per user to generate. For example, if the
        dataset 'a' was generated from a leave-one-out split, and n_samples=100, that
        user would receive 101 samples.

    Returns
    -------
    output: (ndarray, List[ndarray])
        The first element corresponds to an array of _ordered_ user ids, the second
        the per-user samples.

    References
    ----------
    [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569

    """

    users, items, _ = a.to_components(
        negative_samples=n_samples,
        aux_matrix=b.interactions.tocsr(),
        shuffle=False,
        sampling_mode="absolute",
    )

    unique_users = unique(users)

    sampled_users, sampled_items = (
        users[len(unique_users) :],
        items[len(unique_users) :],
    )

    groups, grouped = groupby(sampled_users, sampled_items)

    grouped = c_[items[: len(unique_users)], grouped]

    return unique_users, grouped



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from typing import Optional, Any, List
from itertools import islice

from numpy import asarray

from implicit.als import AlternatingLeastSquares
from implicit.bpr import BayesianPersonalizedRanking

from xanthus.datasets import Dataset


class MatrixFactorizationModel:
    """
    A simple adapter for 'non-neural' matrix factorization algorithms.

    This class wraps the functionality provided by the Implicit library [1] which is
    itself based (partly) upon [2]. It is provided as a baseline model, along with
     the simple 'PopRankModel'.

    References
    ----------
    [1] https://github.com/benfred/implicit/tree/master/implicit
    [2] https://www.comp.nus.edu.sg/~xiangnan/papers/sigir16-eals-cm.pdf

    See Also
    --------
    xanthus.models.baseline.PopRankModel

    """

    _methods = {
        "als": AlternatingLeastSquares,
        "bpr": BayesianPersonalizedRanking,
    }

    def __init__(self, method: str = "als", **kwargs: Optional[Any]) -> None:
        """Initialise a MatrixFactorizationModel."""

        self._model = self._methods[method](**kwargs)
        self._mat = None
        self._method = method

    def fit(self, dataset: Dataset) -> "MatrixFactorizationModel":
        """
        Fit the model to a provided Dataset.

        Parameters
        ----------
        dataset: Dataset
            An input dataset.

        Returns
        -------
        output: MatrixFactorizationModel
            Returns itself. How fun.

        See Also
        --------
        xanthus.datasets.Dataset

        """

        self._mat = dataset.interactions.tocsr()
        self._model.fit(self._mat.T)
        return self

    def predict(
        self,
        dataset: Dataset,
        users: Optional[List[int]] = None,
        items: Optional[List[int]] = None,
        n: int = 3,
        **kwargs: Optional[Any],
    ) -> List[List[int]]:
        """
        Generate predictions (recommendations) from the model. For each provided user,
        the output will be a set of items ranked in order of predicted preference.

        Parameters
        ----------
        dataset: Dataset, optional
            The dataset for which you wish to generate recommendations. If this is
            provided, this object's `all_items` and `users` will be used for the
            purpose of generating recommendations.
        users: list, optional
            An optional array of users for whom you wish to generate recommendations.
        items: list, optional
            An optional array of items you wish to be used in recommendations. This
            may be a subset of items for the purposes of ranking specific subsets of
            items, for example. Maybe you want to see which Star Wars movies are the
            most loved according to the model, for example.
        n: int
            The number of recommendations to be generated per user.
        Returns
        -------
        output: list
            A list, where each element corresponds to a list of recommendations.

        """

        users = users if users is not None else dataset.users
        users = users.flatten()

        user_items = dataset.interactions.tocsr()

        if items is None:
            rec = self._model.recommend
            recommended = asarray(
                [[_[0] for _ in rec(user, user_items, N=n, **kwargs)] for user in users]
            )

        else:
            user_factors = self._model.user_factors
            item_factors = self._model.item_factors
            recommended = []
            for i, user in enumerate(users):
                user_factor = user_factors[user]
                user_items = items[i].flatten()
                scores = item_factors.dot(user_factor)
                user_recs = islice(
                    sorted(zip(user_items, scores[user_items]), key=lambda _: -_[1]), n
                )
                recommended.append([_[0] for _ in user_recs])

        return recommended

    def __repr__(self):
        return f"MatrixFactorizationModel(method={self._method})"



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from typing import Optional, Any, List, Dict, Callable

import numpy as np
from numpy import ndarray
from pandas import DataFrame


class DatasetEncoder:
    """
    A class for simply encoding and decoding recommendation datasets in the spirit of
    a Scikit-Learn Transformer (but not _actually_ a transformer!).

    The objective of this class is to help you manage, save and load your encodings, and
    to give you a few little utilities to help make building a recommendation model that
    bit simpler.

    Examples
    --------

    >>> from xanthus.datasets import DatasetEncoder
    >>> users = ["jane.smith@email.com", "john.appleseed@email.com"]
    >>> items = ["action-movie-0001", "action-movie-0002"]
    >>> encoder = DatasetEncoder()
    >>> encoded = encoder.fit_transform(users=users, items=items)
    >>> output = encoder.to_df(encoded["users"],
    ...                        [encoded["items"] for _ in encoded["users"]])
    >>> print(output)
                             id             item_0             item_1
    0      jane.smith@email.com  action-movie-0001  action-movie-0002
    1  john.appleseed@email.com  action-movie-0001  action-movie-0002

    """

    def __init__(self) -> None:
        """
        Initialize the DatasetEncoder.
        """

        self._offset = 1
        self.user_mapping: Dict[str, int] = {}
        self.item_mapping: Dict[str, int] = {}
        self.user_tag_mapping: Dict[str, int] = {}
        self.item_tag_mapping: Dict[str, int] = {}

    def fit(self, *args: Optional[Any], **kwargs: Optional[Any]) -> "DatasetEncoder":
        """Fit the DatasetEncoder."""

        return self.partial_fit(*args, **kwargs)

    def transform(
        self,
        users: Optional[List[str]] = None,
        items: Optional[List[str]] = None,
        user_tags: Optional[List[str]] = None,
        item_tags: Optional[List[str]] = None,
    ) -> Dict[str, ndarray]:
        """
        Transform a given set of users, items or their related metadata into a set of
        encodings.

        Parameters
        ----------
        users: list
            A list of user IDs to transform into encodings.
        items: list
            A list of item IDs to transform into encodings.
        user_tags: list
            A list of item meta data to transform into encodings.
        item_tags: list
            A list of item meta data to transform into encodings.

        Returns
        -------
        output: dict
            A dictionary where each input (users, items, user_tags, item_tags)
            that has been passed to the method is returned as a key mapped to an array
            of encoded values. The order of input values is preserved.

        Notes
        -----
        * The elements in users/items and user_tags/item_tags need not be aligned for
          this method to work (i.e. the list of users and items need not be valid
          user item pairs in the input set). Indeed, you can encode only users, only
          items, their metadata or any combination of the set.

        """

        output = dict()

        if users is not None:
            output["users"] = np.fromiter(
                (self.user_mapping[_] for _ in users), np.int32
            )

        if items is not None:
            output["items"] = np.fromiter(
                (self.item_mapping[_] for _ in items), np.int32
            )

        if user_tags is not None:
            output["user_tags"] = np.fromiter(
                (self.user_tag_mapping[_] for _ in user_tags), np.int32
            )

        if item_tags is not None:
            output["item_tags"] = np.fromiter(
                (self.item_tag_mapping[_] for _ in item_tags), np.int32
            )

        return output

    def fit_transform(
        self, *args: Optional[Any], **kwargs: Optional[Any]
    ) -> Dict[str, ndarray]:
        """Fit the DatasetEncoder, then transform it. Simples."""

        self.fit(*args, **kwargs)
        return self.transform(*args, **kwargs)

    def partial_fit(
        self,
        users: Optional[List[str]] = None,
        items: Optional[List[str]] = None,
        user_tags: Optional[List[str]] = None,
        item_tags: Optional[List[str]] = None,
    ) -> "DatasetEncoder":
        """
        Fit the DatasetEncoder.

        This will update encodings to include any new user/item values.

        Parameters
        ----------
        users: list, optional
            A list of users to encode.
        items: list, optional
            A list of items to encode.
        user_tags: list, optional
            A list of user features to encode.
        item_tags: list, optional
            A list of item features to encode.

        Returns
        -------
        output: DatasetEncoder
            A newly updated DatasetEncoder.

        """

        if users is not None:
            self.user_mapping = self._fit_mapping(users, self.user_mapping)

        if items is not None:
            self.item_mapping = self._fit_mapping(items, self.item_mapping)

        if user_tags is not None:
            self.user_tag_mapping = self._fit_feature_mapping(
                user_tags, self.user_tag_mapping, self.user_mapping
            )

        if item_tags is not None:
            self.item_tag_mapping = self._fit_feature_mapping(
                item_tags, self.item_tag_mapping, self.item_mapping
            )

        return self

    def inverse_transform(
        self,
        users: Optional[List[int]] = None,
        items: Optional[List[int]] = None,
        user_tags: Optional[List[int]] = None,
        item_tags: Optional[List[int]] = None,
    ) -> Dict[str, ndarray]:
        """
        Given a set of encoded inputs, return their original values.

        Parameters
        ----------
        users: list, optional
            A list of users to decode.
        items: list, optional
            A list of items to decode.
        user_tags: list, optional
            A list of user features to decode.
        item_tags: list, optional
            A list of item features to decode.

        Returns
        -------
        output: dict
            A dictionary where each encoded input (users, items, user_tags, item_tags)
            that has been passed to the method is returned as a key mapped to an array
            of decoded values. The order of input values is preserved.

        """

        output = dict()

        if users is not None:
            inv_user_mappings = {v: k for k, v in self.user_mapping.items()}
            output["users"] = np.asarray(list(inv_user_mappings[_] for _ in users))

        if items is not None:
            inv_item_mappings = {v: k for k, v in self.item_mapping.items()}
            output["items"] = np.asarray(list(inv_item_mappings[_] for _ in items))

        if user_tags is not None:
            inv_user_feat_mappings = {v: k for k, v in self.user_tag_mapping.items()}
            output["user_tags"] = np.asarray(
                list(inv_user_feat_mappings[_] for _ in user_tags)
            )

        if item_tags is not None:
            inv_item_feat_mappings = {v: k for k, v in self.item_tag_mapping.items()}
            output["item_tags"] = np.asarray(
                list(inv_item_feat_mappings[_] for _ in item_tags)
            )

        return output

    def get_params(self) -> Dict[str, Dict[str, int]]:
        """Get the encodings of the current DatasetEncoder."""

        return dict(
            user_mapping=self.user_mapping,
            item_mapping=self.item_mapping,
            user_tag_mapping=self.user_tag_mapping,
            item_tag_mapping=self.item_tag_mapping,
        )

    def set_params(self, params: Dict[str, Dict[str, int]]) -> "DatasetEncoder":
        """Set the encodings to one of your own. What could possibly go wrong?!"""

        self.user_mapping = params.get("user_mapping", self.user_mapping)
        self.item_mapping = params.get("item_mapping", self.item_mapping)
        self.user_tag_mapping = params.get("user_tag_mapping", self.user_tag_mapping)
        self.item_tag_mapping = params.get("item_tag_mapping", self.item_tag_mapping)
        return self

    def to_df(
        self,
        targets: List[int],
        items: List[List[int]],
        mode: str = "users",
        target_col: str = "id",
        item_col_fmt: str = "item_{0}",
        transform: Callable[[List[int]], List[str]] = None,
    ) -> DataFrame:
        """
        Transform a list of users and associated items into a Pandas DataFrame.

        This method is intended to act as a utility for converting *encoded*
        user-recommendation sets into DataFrame format for subsequent post-processing
        and analysis. It is also intended to support item-item 'recommendations' too
        (hence use of 'mode', and 'targets').

        Parameters
        ----------
        targets: list
            A list of encoded targets (e.g. users).
        items: list
            A list, where each element is a list of encoded items mapped to a given
            user (i.e. the user at {i} corresponds to the set of items at {i} in this
            list.
        mode: str
            The mode to use when creating the output frame. Either 'users' or 'items'.
            In the former case, 'targets' will be treated as users, in the latter, as
            items.
        target_col: str
            The name of the output column associated with decoded targets.
        item_col_fmt: str
            The format of the field names associated with items.
        transform: callable, optional
            An optional callable object to be used to transform encodings. If not
            provided, the `inverse_transform` method will be used. This can be
            overridden to postprocess decoded items (e.g. to map to SKUs or descriptive
            names).

        Returns
        -------
        output: DataFrame
            A DataFrame containing _decoded_ users and items associated with them.

        Notes
        -----
        * The item list should correspond to items associated with each of the
          users in the input list. For example, this could be one or more items
          recommended to that users.

        """

        if len(targets) != len(items):
            raise ValueError(
                f"The total number of users ('{len(targets)}') does not match the total "
                f"number of item recommendation rows ('{len(items)}')."
            )

        # create column headers.
        n = len(items[0])
        transform = transform or self.inverse_transform
        columns = [target_col, *(item_col_fmt.format(i) for i in range(n))]

        # inverse-transform user/item arrays.
        if mode == "users":
            targets = transform(users=targets)["users"]
        else:
            targets = transform(items=targets)["items"]

        items = [transform(items=items[i])["items"] for i in range(len(items))]

        # stack users and items.
        data = np.c_[np.asarray(targets), np.asarray(items)]

        return DataFrame(data=data, columns=columns)

    def _fit_feature_mapping(
        self, x: List[str], encodings: Dict[str, int], aux_encodings: Dict[str, int]
    ) -> Dict[str, int]:
        """Generate feature encodings for a given decoded list of features 'x'."""

        encodings = self._fit_mapping(x, encodings, offset=len(aux_encodings),)
        encodings.update(aux_encodings)
        return encodings

    def _fit_mapping(
        self, x: List[str], encodings: Dict[str, int], offset: int = 0
    ) -> Dict[str, int]:
        """Generate encodings for a given decoded list 'x'."""

        for e in x:
            encodings.setdefault(e, len(encodings) + offset + self._offset)
        return encodings



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from typing import Optional, Union, List, Set, Iterator, Tuple, Any, Callable
from itertools import islice

from pandas import DataFrame
from scipy.sparse import coo_matrix, csr_matrix, dok_matrix
import numpy as np
from numpy import random, ones_like, ndarray, zeros


# this is a bit of a mess.
SamplerCallable = Callable[
    [ndarray, ndarray, ndarray, csr_matrix, int, str, csr_matrix, bool],
    Tuple[ndarray, ndarray, ndarray],
]


def construct_coo_matrix(
    p: ndarray,
    q: ndarray,
    r: Optional[ndarray] = None,
    dtype: str = "float32",
    shape: Optional[Tuple[int, int]] = None,
) -> coo_matrix:
    """Build a COO matrix."""

    r = r if r is not None else ones_like(p)
    return coo_matrix((r.astype(dtype), (p, q)), shape=shape)


def construct_dok_matrix(
    p: ndarray,
    q: ndarray,
    r: Optional[ndarray] = None,
    dtype: str = "float32",
    shape: Optional[Tuple[int, int]] = None,
) -> dok_matrix:
    """Build a DOK matrix."""

    if shape is None:
        shape = (p.max() + 1, q.max() + 1)

    r = r if r is not None else ones_like(p)

    mat = dok_matrix(shape, dtype=dtype)

    for i, k in enumerate(zip(p, q)):
        mat[k] = r[i]

    return mat


def rejection_sample(p: int, q: Union[List, Set]) -> Iterator[int]:
    """
    Sample without replacement from a set of 'p' items that are not in set 'q'.

    This function will continue to draw from 'p' until all elements in 'p' are in 'q'.

    Parameters
    ----------
    p: int
        The number of elements to sample from.
    q: set-like
        An set of items to excluded from sampling, e.g. existing purchases.

    Returns
    -------
    output: Iterator
        A generator yielding integers, where each integer corresponds to a negative
        sample (i.e. a sample from 'p' that is not in 'q').

    """

    sampled = set(q).copy()
    sampled.add(0)

    while len(sampled) < p:
        choice = random.randint(p)
        if choice not in sampled:
            yield choice
            sampled.add(choice)


def relative_rejection_sample(p: int, q: Union[List, Set], k: int,) -> Iterator[int]:
    """
    Sample without replacement from a set of 'p' items that are not in set 'q', where
    the number of samples 'k' is drawn for each element in the set 'q'

    Parameters
    ----------
    p: int
        The number of elements to sample from.
    q: set-like
        An set of items to excluded from sampling, e.g. existing purchases.
    k: int
        The number of samples to draw from 'p' _for each_ element in 'q'.

    Returns
    -------
    output: Iterator
        A generator yielding integers, where each integer corresponds to a negative
        sample (i.e. a sample from 'p' that is not in 'q').

    """

    yield from (a for _ in range(len(q)) for a in islice(rejection_sample(p, q), k))


def single_negative_sample(
    i: int, mat: csr_matrix, k: int, mode: str = "relative"
) -> Iterator[int]:
    """
    Generate single negative sample of at least 'k' elements for user index 'i' in
    interaction matrix 'mat'.

    Parameters
    ----------
    i: int
        The index (in the 'mat') of the 'user' for whom we wish to generate negative
        samples.
    mat: csr_matrix
        A sparse matrix containing interaction mappings (e.g. non-zero elements
        correspond to an user-item interaction, perhaps where a user has purchased or
        rated an item).
    k: int
        The number of samples to draw from items in 'mat'. If mode is 'relative', this
        will sample 'mat' _for each_ element in 'mat' that is non-zero.
    mode: str
        One of two modes:
            'relative' - Sample 'k' negative items from 'p' (total items), _for each_
                         item the user 'i' has interacted with (i.e. for each positive,
                         sample 'k' negatives). This will produce 'k' x 'p' samples.
            'absolute' - Sample 'k' negative items from 'p' (total items). This will
                         produce 'k' samples.

    Returns
    -------
    output: Iterator
        A generator yielding integers, where each integer corresponds to a negative
        sample (i.e. a sample from 'p' that is not in 'q').

    """

    n = mat.shape[1]
    sampled = mat[i].nonzero()[1]

    if mode == "relative":
        yield from relative_rejection_sample(n, sampled, k)
    elif mode == "absolute":
        yield from islice(rejection_sample(n, sampled), k)
    else:
        raise ValueError(f"Unknown negative sampling mode '{mode}'.")


def iter_negative_samples(
    arr: List[int], mat: csr_matrix, k: int, **kwargs: Optional[Any]
) -> Iterator[Tuple[int, int]]:
    """
    Generate single negative samples of at least 'k' elements for each user id 'i' in
    interaction matrix 'mat'.

    Parameters
    ----------
    arr: list
        A list of corresponding ids.
    mat: csr_matrix
        A sparse matrix containing interaction mappings (e.g. non-zero elements
        correspond to an user-item interaction, perhaps where a user has purchased or
        rated an item).
    k: int
        The number of samples to draw from items in 'mat'. If mode is 'relative', this
        will sample 'mat' _for each_ element in 'mat' that is non-zero.
    kwargs: any, optional
        See `single_negative_sample`.

    Returns
    -------
    output: Iterator
        An iterator, where each element corresponds to a negative user-item pair.

    Notes
    -----
    * It's assumed that the row-index of the sparse matrix corresponds to users, while
      the column-index of the matrix corresponds to items, i.e.:
        mat.shape == (n_users, n_items)

    See Also
    --------
    xanthus.dataset.utils.single_negative_sample

    """
    yield from (
        (i, z) for i in arr for z in single_negative_sample(i, mat, k, **kwargs)
    )


def unpack_negative_samples(
    arr: ndarray, mat: csr_matrix, k: int, **kwargs: Optional[Any]
) -> Tuple[ndarray, ndarray, ndarray]:
    """
    Generate single negative samples of at least 'k' elements for each user id 'i' in
    interaction matrix 'mat' _as a set of ndarrays_.

    This is a utility function wrapping the behaviour of `iter_negative_samples`.

    Parameters
    ----------
    arr: list
        A list of corresponding ids.
    mat: csr_matrix
        A sparse matrix containing interaction mappings (e.g. non-zero elements
        correspond to an user-item interaction, perhaps where a user has purchased or
        rated an item).
    k: int
        The number of samples to draw from items in 'mat'. If mode is 'relative', this
        will sample 'mat' _for each_ element in 'mat' that is non-zero.
    kwargs: any, optional
        See `single_negative_sample`.

    Returns
    -------
    output: tuple
        A set of three arrays, with the first typically corresponding to user ids, the
        second to item ids, and the third to ratings.

    See Also
    --------
    iter_negative_samples

    """

    arr = list(iter_negative_samples(np.unique(arr), mat, k, **kwargs))
    data = np.asarray(arr).astype(np.int32)
    return data[:, 0], data[:, 1], zeros(shape=data.shape[0])


def sample_negatives(
    users: ndarray,
    items: ndarray,
    ratings: ndarray,
    interactions: csr_matrix,
    negative_samples: int,
    sampling_mode: str = "absolute",
    aux_matrix: Optional[csr_matrix] = None,
    concat: bool = True,
) -> Tuple[ndarray, ndarray, ndarray]:
    """
    Draw negative samples from the provided dataset, and optionally return these
    samples concatenated to the provided user/item/rating arrays.

    For the purposes of this function, a negative sample is a sample of an _item_ for
    a _user_ that that user has not interacted with (given the provided parameters).

    Parameters
    ----------
    users: ndarray
        An array containing user IDs.
    items: ndarray
        An array containing item IDs associated with a given user ID.
    ratings: ndarray
        An array corresponding to ratings for a given user-item interaction.
    interactions: csr_matrix
        The sparse matrix storing interactions info.
    negative_samples: int
        The number of negative samples to draw.
    sampling_mode: str
        One of two modes:
            'relative' - Sample 'k' negative items from 'p' (total items), _for each_
                         item the user 'i' has interacted with (i.e. for each positive,
                         sample 'k' negatives). This will produce 'k' x 'p' samples.
            'absolute' - Sample 'k' negative items from 'p' (total items). This will
                         produce 'k' samples.
    aux_matrix: csr_matrix, optional
        Optionally provide a second CSR matrix. This may contain additional interactions
        you wish to utilise in the sampling, but may not be present in the original set.
    concat: bool
        Indicate whether you wish the negative samples to be returned concatenated to
        your input user/item/rating arrays.

    Returns
    -------
    output: tuple
        A set of three arrays, with the first typically corresponding to user ids, the
        second to item ids, and the third to ratings.

    """

    if aux_matrix is not None:
        interactions += aux_matrix

    neg_users, neg_items, neg_ratings = unpack_negative_samples(
        users, interactions, negative_samples, mode=sampling_mode
    )

    if concat:
        users = np.concatenate((users, neg_users))
        items = np.concatenate((items, neg_items))
        ratings = np.concatenate((ratings, neg_ratings))

        return users, items, ratings
    else:
        return neg_users, neg_items, neg_ratings


def groupby(key: ndarray, values: ndarray) -> Tuple[ndarray, List[ndarray]]:
    """
    Execute a simple pure-numpy group-by operation over two arrays.

    Note that both arrays should be of the same type.

    Parameters
    ----------
    key: ndarray
        The array containing keys (values you wish to group on). Assumed to be 1D.
    values: ndarray
        The array containing values (that you wish to group together). Assumed to be 1D.

    Returns
    -------
    output: tuple
        A pair, where the first element is the ordered group keys, and the second a list
        of ndarrays such that each element in order maps to a group key at the same
        index, such that dict(zip(groups, grouped)) would produce a key-value map to the
        groups.

    """

    x = np.c_[key, values]
    idx = np.argsort(x[:, 0])
    x = x[idx]

    groups, counts = np.unique(x[:, 0], return_counts=True)
    grouped = np.split(x[:, 1], np.cumsum(counts)[:-1])

    return groups, grouped


def as_implicit(a: ndarray) -> ndarray:
    """
    Utility method for casting all scores as implicit (i.e. all ones).

    In practice, a sugary alias for `lambda _: ones_like(_)`.

    Parameters
    ----------
    a: ndarray
        An array of floats corresponding to user-item scores.

    Returns
    -------
    output: ndarray
        An array of ones. Simple, really.

    """

    return np.ones_like(a)


def fold(
    df: DataFrame,
    key: str,
    cols: List[str],
    fn: Optional[Callable[[str], Iterator[str]]] = None,
    deduplicate: bool = True,
) -> DataFrame:
    """
    'Fold' a wide DataFrame into a DataFrame in the correct format to be passed to a
    xanthus Dataset as metadata (i.e. a DataFrame with two columns, '{key}' and 'tag').

    Parameters
    ----------
    df: DataFrame
        An input dataframe of the format:

            key|cols{0}|...|cols{n}
            :-:|:-----:|:-:|:-----:
            ...

        Where 'key' and 'cols{0}, ..., cols{n}' correspond to the 'key' and 'cols'
        parameter values. See the example below for more details.
    key: str
        The primary key (e.g. unique customer ID) of the input DataFrame.
    cols: list
        A list of column names to be 'folded' (i.e. metadata fields).
    fn: callable, optional
        An optional callable that can be used, for example, to convert a cell in a
        DataFrame into a set of tokens. This can be used on the MovieLens dataset to
        unpack movie genres, for example.
    deduplicate: bool
        Deduplicate the resulting frame to ensure all key-tag pairs occur exactly once.

    Returns
    -------
    output: DataFrame
        A 'folded' DataFrame ready to be used as input to a Dataset

    Examples
    --------

    >>> import pandas as pd
    >>> data = [["jane smith", "london", "doctor"],
    ...         ["dave davey", "manchester", "spaceman"],
    ...         ["john appleseed", "san francisco", "corporate shill"],
    ...         ["jenny boo", "paris", "ninja"]]
    >>> raw_meta = pd.DataFrame(data=data,
    ...                            columns=["user", "location", "occupation"])
    >>> meta = fold(raw_meta, "user", ["location", "occupation"])
    >>> meta
                 user              tag
    0      jane smith           london
    1      jane smith           doctor
    2      dave davey       manchester
    3      dave davey         spaceman
    4  john appleseed    san francisco
    5  john appleseed  corporate shill
    6       jenny boo            paris
    7       jenny boo            ninja

    """

    tags = df[cols].values.tolist()
    keys = df[key].values.tolist()

    if fn is None:
        pairs = ((k, t) for i, k in enumerate(keys) for t in tags[i])
    else:
        pairs = (
            (k, t)
            for i, k in enumerate(keys)
            for t in (e for _ in tags[i] for e in fn(_))
        )

    output = DataFrame(data=pairs, columns=[key, "tag"])

    return output.drop_duplicates([key, "tag"])



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from typing import Optional, Any, List, Tuple, Callable, Iterator
from collections import defaultdict

from numpy import ndarray
from pandas import DataFrame
from scipy.sparse import coo_matrix, csr_matrix
from sklearn.utils import shuffle as _shuffle

import numpy as np
from .encoder import DatasetEncoder


from .utils import construct_coo_matrix, sample_negatives, SamplerCallable


class Dataset:
    """
    A simple dataset abstraction with utilities for training and evaluating
    recommendation models. It was designed to help implement the methodology
    outlined in He et al. (i.e. negative sampling), alongside extensions mention in said
    paper (using user- & item-metadata with implicit feedback).

    With this class you can:
    * Lazily generate user-item vectors for model training and evaluation.
    * Lazily map users and items to user and item metadata.
    * Generate negative samples. This can be used for training data and to create
      and to build evaluation data sets to test model ranking performance.
    * Easily generate the above directly from a Pandas DataFrame.

    In most cases, you'll find it easiest to simply call the Dataset.from_df method,
    below.

    Parameters
    ----------
    interactions: coo_matrix
        A user-item interaction matrix of the shape (NxM), where N is the number of
        users, and M is the number of items.
    user_meta: coo_matrix
        A user-user_metadata matrix of the shape (Nx(N+K)), where N is the number of
        users, and K is the number of unique tags associated with users.
    item_meta: coo_matrix
        A item-item_metadata matrix of the shape (Nx(N+K)), where N is the number of
        items, and K is the number of unique tags associated with items.
    encoder: DatasetEncoder
        A _fitted_ DatasetEncoder object. This should have been fitted on the same data
        that was used to generate the `interaction`, `user_meta` and `item_meta`
        matrices. Note that calling the `from_df` classmethod will automatically
        initialise and fit an encoder for you.
    sampler: SamplerCallable
        This is a callable that is used to sample from your interactions data. A simple
        negative sampler is provided, but you should feel free to experiment with
        your own.

    Notes
    -----
    * Tags are user- or item-metadata. For example, for movie metadata, for the movie
      'Lord of the Rings: Fellowship of the Ring', the movie may have the ID '1', and
      may have the tags: 'fantasy', 'adventure'. If these tags had the encodings '2' and
      '3' respectively, then the elements (1, 2) and (1, 3) in the item metadata matrix
      would be non-zero values.
    * This class is not suitable for very large datasets, it could do with a bit of TLC
      to make it more friendly with large use-cases.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.read_csv("data/movielens-100k/ratings.csv")
    >>> df = df.rename(columns={"userId": "user", "movieId": "item"})
    >>> ds = Dataset.from_df(df)
    >>> for user, item, rating in ds.iter(shuffle=False):
    ...    print(user, item, rating)
    ...    break
    ...
    [1] [1] 4.0

    >>> df = pd.read_csv("data/movielens-100k/ratings.csv")
    >>> df = df.rename(columns={"userId": "user", "movieId": "item"})
    >>> metadata = pd.read_csv("data/movielens-100k/movies.csv")
    >>> item = metadata["movieId"]
    >>> tags = metadata["genres"]
    >>> data = ([item[i], tag] for i in range(len(item)) for tag in tags[i].split("|"))
    >>> item_meta = pd.DataFrame(data=data, columns=["item", "tag"])
    >>> ds = Dataset.from_df(df, item_meta=item_meta)
    >>> for user, item, rating in ds.iter(output_dim=4, shuffle=False):
    ...     print(user, item, rating)
    ...     break
    [1] [1, 9743, 9744, 9745] 4.0

    """

    def __init__(
        self,
        interactions: coo_matrix,
        user_meta: coo_matrix,
        item_meta: coo_matrix,
        encoder: DatasetEncoder,
        sampler: SamplerCallable = sample_negatives,
    ) -> None:
        """
        Initialise a Dataset.
        """

        self.encoder = encoder
        self.sampler = sampler

        self.interactions = interactions
        self.user_meta = user_meta
        self.item_meta = item_meta

    @property
    def users(self) -> ndarray:
        """Get all users _with interactions_ in the dataset."""

        return np.unique(self.interactions.nonzero()[0])

    @property
    def items(self) -> ndarray:
        """Get all items _with interactions_ in the dataset."""

        return np.unique(self.interactions.nonzero()[1])

    @property
    def all_users(self) -> ndarray:
        """Get all users in the dataset (inc. those with no interactions)."""

        return np.arange(self.interactions.shape[0])

    @property
    def all_items(self) -> ndarray:
        """Get all items in the dataset (inc. those with no interactions)."""

        return np.arange(self.interactions.shape[1])

    @property
    def history(self) -> ndarray:
        """Get the history (items a user has interacted with) of each user."""

        mat = self.interactions.tocsr()
        g = (mat[i].nonzero()[1].tolist() for i in self.users)
        return np.fromiter(g, dtype=int)

    def iter(
        self,
        negative_samples: int = 0,
        output_dim: int = 1,
        shuffle: bool = True,
        aux_matrix: Optional[coo_matrix] = None,
        sampling_mode: str = "relative",
    ) -> Iterator[Tuple[ndarray, ndarray, float]]:
        """
        Iterate over a sequence of ([user_vector_{i}, item_vector_{i}], ratings_{i}).

        In practice, this will result in each user-item interaction being yielded,
        optionally with additional metadata, and optionally with 'n' negative sample
        instances.

        By default (with 'output_dim=1', and/or with no metadata), this will yield:

        ([user_{i}], [item_{i}}], ratings_{i})

        Concretely, this may be: ([0], [1], 1) If user/item meta-data is provided, this
        will be lazily injected into the yielded value, for example:

        ([user_{i}, user_tag_{i}{1}, ..., user_tag_{i}{n}],
         [item_{i}, item_tag_{i}{1}, ..., item_tag_{i}{n}],
         ratings_{i})

        Again, concretely: ([0, 21, 82], [1, 97, 64], 1). Make sure to set 'output_dim'
        if you want your metadata rendered (if you provided it)!

        Parameters
        ----------
        negative_samples: int
            The total number of negative samples (for each positive sample) you wish
            to take from the provided interactions set (and auxiliary matrix, if
            provided).
        output_dim: int
            The output dimensions for _both_ the encoded user- and item-vectors. Note
            that this will only be applied if user/item metadata is provided, otherwise
            all output vectors will have 'output_dim=1'.
        shuffle: int
            Indicate whether the output data should be shuffled.
        aux_matrix: coo_matrix, optional
            Provide a sparse matrix of the same shape as the 'interactions' matrix with
            additional interactions terms. These terms will be used when taking negative
            samples. This can be useful if this Dataset is a 'test' dataset, and you
            wish to draw negative samples from items a user has never interacted with
            when generating an evaluation set. This is the process described in [1].
        sampling_mode: str
            If negative sampling is used, specify the sampling mode you wish to use.
            See 'xanthus.dataset.utils.single_negative_sample' for more details.

        Returns
        -------
        output: Generator
            A generator yielding user/item vectors and the associated pairing's rating.

        See Also
        --------
            xanthus.evaluate.utils.he_sample
            xanthus.dataset.utils.single_negative_sample

        References
        ----------
        [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569

        """

        # must cast interactions to csr so we can use indexing on the matrix.
        interactions: csr_matrix = self.interactions.tocsr()

        # setup user metadata
        if self.user_meta is not None:
            user_meta = self.user_meta.tocsr()
        else:
            user_meta = None

        # setup item metadata
        if self.item_meta is not None:
            item_meta = self.item_meta.tocsr()
        else:
            item_meta = None

        users, items = interactions.nonzero()
        ratings = interactions.data

        if negative_samples > 0:
            # the aux_matrix should include additional interactions you wish to consider
            # _exclusively_ for the purposes of generating negative samples.
            users, items, ratings = self.sampler(
                users,
                items,
                ratings,
                interactions,
                negative_samples,
                sampling_mode,
                aux_matrix,
                concat=True,
            )

        # optionally shuffle the users, items and ratings.
        if shuffle:
            users, items, ratings = _shuffle(users, items, ratings)

        ratings.reshape(-1, 1)

        # stack user ids with associated user metadata.
        if user_meta is not None and output_dim > 1:
            users = self._iter_meta(users, user_meta, output_dim)
        elif output_dim > 1:
            users = np.c_[users, np.zeros((len(users), output_dim - 1), dtype=int)]
        else:
            users = users.reshape(-1, 1)

        # stack item ids with associated item metadata.
        if item_meta is not None and output_dim > 1:
            items = self._iter_meta(items, item_meta, output_dim)
        elif output_dim > 1:
            items = np.c_[items, np.zeros((len(items), output_dim - 1))]
        else:
            items = items.reshape(-1, 1)

        for (user, item, rating) in zip(users, items, ratings):
            yield user, item, rating

    def iter_user(self, users: ndarray, n_dim: int = 1) -> Iterator[ndarray]:
        """Iterate over user metadata vectors of size 'n_dim'."""

        yield from self._iter_ids(users, self.user_meta, n_dim)

    def iter_item(self, items: ndarray, n_dim: int = 1) -> Iterator[ndarray]:
        """Iterate over item metadata vectors of size 'n_dim'."""

        yield from self._iter_ids(items, self.item_meta, n_dim)

    def _iter_ids(self, ids: ndarray, mat: coo_matrix, n_dim: int) -> Iterator[ndarray]:
        """
        Iterate over metadata vectors of size 'n_dim' encoded in 'mat'.

        If too few metadata elements are found to satisfy 'n_dim' requirement, the
        vectors will be padded with zeros to ensure homogeneity.

        See Also
        --------
        Dataset._iter_meta

        """

        if mat is not None:
            yield from self._iter_meta(ids, mat.tocsr(), n_dim)
        elif n_dim > 1:
            ids = np.c_[ids, np.zeros((len(ids), n_dim - 1), dtype=int)]
            yield from (_ for _ in ids)
        else:
            yield from (_ for _ in ids.reshape(-1, 1))

    @staticmethod
    def _iter_meta(ids: ndarray, meta: csr_matrix, n_dim: int) -> Iterator[List[int]]:
        """
        Lazily evaluate metadata in the provided CSR matrix.

        Parameters
        ----------
        ids: ndarray
            An array of IDs. For items, this will correspond to individual item IDs.
            For users, this will correspond to individual user IDs.
        meta: csr_matrix
            A sparse matrix of (NxM) dimensions, where N corresponds to the number of
            user/item IDs (above) and M corresponds to the number of user/item metadata
            features (vocab) in the dataset.
        n_dim: int
            The length of the output vectors. Makes sure this is large enough to
            actually append some metadata to your output vectors (i.e. > 1).

        Returns
        -------
        output: Iterator
            An iterator, where each ID in the list is mapped to corresponding metadata.
            The output shape of each element is then a list of 'n_dim' length.

        """

        groups = defaultdict(list)
        _ids, tags = meta.nonzero()

        for _id, _tag in zip(_ids, tags):
            groups[_id].append(_tag)

        for _id in ids:
            group = groups[_id]
            padding = [0] * max(0, n_dim - len(group))
            features = [_id, *group, *padding][:n_dim]
            yield features

    @classmethod
    def from_df(
        cls,
        interactions: DataFrame,
        user_meta: Optional[DataFrame] = None,
        item_meta: Optional[DataFrame] = None,
        normalize: Optional[Callable[[ndarray], ndarray]] = None,
        encoder: Optional[DatasetEncoder] = None,
        **kwargs: Optional[Any]
    ) -> "Dataset":
        """
        Initialise the Dataset directly from an interactions DataFrame and optionally
        from additional user_meta and item_meta DataFrames too.

        Parameters
        ----------
        interactions: coo_matrix
            A user-item interaction matrix of the shape (NxM), where N is the number of
            users, and M is the number of items.
        user_meta: coo_matrix
            A user-user_metadata matrix of the shape (Nx(N+K)), where N is the number of
            users, and K is the number of unique tags associated with users.
        item_meta: coo_matrix
            A item-item_metadata matrix of the shape (Nx(N+K)), where N is the number of
            items, and K is the number of unique tags associated with items.
        normalize: callable, optional
            Provide a normalization function to apply to your scores. If not provided,
            you'll get what you put in. Simple, right?
        encoder: DatasetEncoder, optional
            Optionally provide a pre-trained DatasetEncoder. This can be useful if you
            want to share the same encoder over two Datasets to preserve the same
            encodings across the datasets.

        Returns
        -------
        output: Dataset
            A shiny new Dataset, ready for your next recommender experiment.

        """

        users = set(interactions["user"].tolist())
        items = set(interactions["item"].tolist())

        if user_meta is not None:
            users.update(set(user_meta["user"].tolist()))

        if item_meta is not None:
            items.update(set(item_meta["item"].tolist()))

        if encoder is None:
            encoder = DatasetEncoder(**kwargs)
            encoder.fit(
                users,
                items,
                user_meta["tag"].unique() if user_meta is not None else None,
                item_meta["tag"].unique() if item_meta is not None else None,
            )

        encoded = encoder.transform(interactions["user"], interactions["item"])

        if "rating" not in interactions.columns:
            ratings = np.ones_like(encoded["users"]).astype(np.float32)
        else:
            ratings = interactions["rating"].values.astype(np.float32)

        if normalize is not None:
            ratings = normalize(ratings)

        interactions_shape = (
            len(encoder.user_mapping) + 1,
            len(encoder.item_mapping) + 1,
        )

        interactions = construct_coo_matrix(
            encoded["users"], encoded["items"], ratings, shape=interactions_shape
        )

        if user_meta is not None:
            user_meta_shape = (
                len(encoder.user_mapping) + 1,
                len(encoder.user_tag_mapping) + 1,
            )
            encoded = encoder.transform(
                users=user_meta["user"], user_tags=user_meta["tag"]
            )
            user_meta = construct_coo_matrix(
                encoded["users"], encoded["user_tags"], shape=user_meta_shape
            )

        if item_meta is not None:
            item_meta_shape = (
                len(encoder.item_mapping) + 1,
                len(encoder.item_tag_mapping) + 1,
            )
            encoded = encoder.transform(
                items=item_meta["item"], item_tags=item_meta["tag"]
            )
            item_meta = construct_coo_matrix(
                encoded["items"], encoded["item_tags"], shape=item_meta_shape
            )

        return cls(
            interactions, user_meta=user_meta, item_meta=item_meta, encoder=encoder
        )

    def to_components(
        self, *args: Optional[Any], **kwargs: Optional[Any]
    ) -> Tuple[ndarray, ...]:
        """Transform `iter` output into a set of arrays."""

        return tuple(map(np.asarray, zip(*self.iter(*args, **kwargs))))

    def __iter__(self) -> Iterator[Tuple[ndarray, ndarray, float]]:
        """Iterate over the dataset."""

        yield from self.iter()



import os
import io
import requests
import zipfile


def movielens(
    version: str = "latest-small",
    base_url: str = "http://files.grouplens.org/datasets/movielens/ml-{version}.zip",
    output_dir: str = "data",
    unzip: bool = True,
) -> None:
    """Download a given movielens dataset."""

    response = requests.get(base_url.format(version=version))

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if unzip:
        zipfile.ZipFile(io.BytesIO(response.content)).extractall(output_dir)
    else:
        with open(os.path.join(output_dir, f"ml-{version}", "wb")) as file:
            file.write(response.content)



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from typing import Optional, Any, Tuple
from tensorflow.keras import Model
from tensorflow.keras.layers import Multiply, Dense, Concatenate
from tensorflow.keras.initializers import lecun_uniform
from tensorflow.keras.regularizers import l2

from xanthus.datasets import Dataset
from xanthus.models import utils, base


class MultiLayerPerceptronModel(base.NeuralRecommenderModel):
    """
    An implementation of a Multilayer Perceptron (MLP) model in Keras.

    Parameters
    ----------
    layers: tuple
        A tuple, where each element corresponds to the number of units in each of the
        layers of the MLP.
    activations: str
        The activation function to use for each of the layers in the MLP.
    l2_reg: float
        The L2 regularization to be applied to each of the layers in the MLP.

    References
    ----------
    [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569

    See Also
    --------
    xanthus.models.base.NeuralRecommenderModel

    """

    def __init__(
        self,
        *args: Optional[Any],
        layers: Tuple[int, ...] = (64, 32, 16, 8),
        activations: str = "relu",
        l2_reg: float = 1e-3,
        **kwargs: Optional[Any]
    ):
        """Initialize a MultiLayerPerceptronModel."""

        super().__init__(*args, **kwargs)
        self._activations = activations
        self._layers = layers
        self._l2_reg = l2_reg

    def _build_model(
        self,
        dataset: Dataset,
        n_user_dim: int = 1,
        n_item_dim: int = 1,
        n_factors: int = 50,
        **kwargs: Optional[Any]
    ) -> Model:
        """
        Build a Keras model, in this case a MultiLayerPerceptronModel (MLP)
        model. See [1] for more info. The original code released with [1] can be
        found at [2].

        Parameters
        ----------
        dataset: Dataset
            The input dataset. This is used to specify the 'vocab' size of each of the
            'embedding blocks' (of which there are two in this architecture).
        n_user_dim: int
            The dimensionality of the user input vector. When using metadata, you should
            make sure to set this to the size of each of these vectors.
        n_item_dim: int
            The dimensionality of the item input vector. When using metadata, you should
            make sure to set this to the size of each of these vectors.
        n_factors: int
            The dimensionality of the latent feature space _for both users and items_
            for the GMF component of the architecture.

        Returns
        -------
        output: Model
            The 'complete' Keras Model object.

        References
        ----------
        [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569
        [2] https://github.com/hexiangnan/neural_collaborative_filtering
        """

        n_user_vocab = dataset.all_users.shape[0]
        n_item_vocab = dataset.all_items.shape[0]

        if dataset.user_meta is not None:
            n_user_vocab += dataset.user_meta.shape[1]
        if dataset.item_meta is not None:
            n_item_vocab += dataset.item_meta.shape[1]

        # mlp block
        user_input, user_bias, user_factors = utils.get_embedding_block(
            n_user_vocab, n_user_dim, int(self._layers[0] / 2)
        )
        item_input, item_bias, item_factors = utils.get_embedding_block(
            n_item_vocab, n_item_dim, int(self._layers[0] / 2)
        )

        body = Concatenate()([user_factors, item_factors])

        for layer in self._layers:
            body = Dense(
                layer,
                activity_regularizer=l2(self._l2_reg),
                activation=self._activations,
            )(body)

        output = Dense(1, activation="sigmoid", kernel_initializer=lecun_uniform())(
            body
        )

        return Model(inputs=[user_input, item_input], outputs=output)


class NeuralMatrixFactorizationModel(base.NeuralRecommenderModel):
    """
    An implementation of a Neural Matrix Factorization (NeuMF) model in Keras.

    Parameters
    ----------
    layers: tuple
        A tuple, where each element corresponds to the number of units in each of the
        layers of the MLP.
    activations: str
        The activation function to use for each of the layers in the MLP.
    l2_reg: float
        The L2 regularization to be applied to each of the layers in the MLP.

    References
    ----------
    [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569

    See Also
    --------
    xanthus.models.base.NeuralRecommenderModel

    """

    def __init__(
        self,
        *args: Optional[Any],
        layers: Tuple[int, ...] = (64, 32, 16, 8),
        activations: str = "relu",
        l2_reg: float = 1e-3,
        **kwargs: Optional[Any]
    ):
        """Initialize a MultiLayerPerceptronModel."""

        super().__init__(*args, **kwargs)
        self._activations = activations
        self._layers = layers
        self._l2_reg = l2_reg

    def _build_model(
        self,
        dataset: Dataset,
        n_user_dim: int = 1,
        n_item_dim: int = 1,
        n_factors: int = 50,
        **kwargs: Optional[Any]
    ) -> Model:
        """
        Build a Keras model, in this case a NeuralMatrixFactorizationModel (NeuMF)
        model. This is a recommender model with two input branches (one half the same
        architecture as in GeneralizedMatrixFactorizationModel, the other the same
        architecture as in MultiLayerPerceptronModel. See [1] for more info. The
        original code released with [1] can be found at [2].

        Parameters
        ----------
        dataset: Dataset
            The input dataset. This is used to specify the 'vocab' size of each of the
            'embedding blocks' (of which there are four in this architecture).
        n_user_dim: int
            The dimensionality of the user input vector. When using metadata, you should
            make sure to set this to the size of each of these vectors.
        n_item_dim: int
            The dimensionality of the item input vector. When using metadata, you should
            make sure to set this to the size of each of these vectors.
        n_factors: int
            The dimensionality of the latent feature space _for both users and items_
            for the GMF component of the architecture.

        Returns
        -------
        output: Model
            The 'complete' Keras Model object.

        References
        ----------
        [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569
        [2] https://github.com/hexiangnan/neural_collaborative_filtering

        """

        n_user_vocab = dataset.all_users.shape[0]
        n_item_vocab = dataset.all_items.shape[0]

        if dataset.user_meta is not None:
            n_user_vocab += dataset.user_meta.shape[1]
        if dataset.item_meta is not None:
            n_item_vocab += dataset.item_meta.shape[1]

        # mlp block
        user_input, mlp_user_bias, mlp_user_factors = utils.get_embedding_block(
            n_user_vocab, n_user_dim, int(self._layers[0] / 2)
        )
        item_input, mlp_item_bias, mlp_item_factors = utils.get_embedding_block(
            n_item_vocab, n_item_dim, int(self._layers[0] / 2)
        )

        mlp_body = Concatenate()([mlp_user_factors, mlp_item_factors])

        for layer in self._layers:
            mlp_body = Dense(
                layer,
                activity_regularizer=l2(self._l2_reg),
                activation=self._activations,
            )(mlp_body)

        # mf block
        user_input, mf_user_bias, mf_user_factors = utils.get_embedding_block(
            n_user_vocab, n_user_dim, n_factors, inputs=user_input,
        )
        item_input, mf_item_bias, mf_item_factors = utils.get_embedding_block(
            n_item_vocab, n_item_dim, n_factors, inputs=item_input,
        )
        mf_body = Multiply()([mf_user_factors, mf_item_factors])

        body = Concatenate()([mf_body, mlp_body])

        output = Dense(1, activation="sigmoid", kernel_initializer=lecun_uniform())(
            body
        )

        return Model(inputs=[user_input, item_input], outputs=output)


class GeneralizedMatrixFactorizationModel(base.NeuralRecommenderModel):
    """
    An implementation of a Generalized Matrix Factorization (GMF) model in Keras.

    References
    ----------
    [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569

    """

    def _build_model(
        self,
        dataset: Dataset,
        n_user_dim: int = 1,
        n_item_dim: int = 1,
        n_factors: int = 50,
        **kwargs: Optional[Any]
    ) -> Model:
        """
        Build a Keras model, in this case a GeneralizedMatrixFactorizationModel (GMF)
        model. See [1] for more info. The original code released with [1] can be
        found at [2].

        Parameters
        ----------
        dataset: Dataset
            The input dataset. This is used to specify the 'vocab' size of each of the
            'embedding blocks' (of which there are two in this architecture).
        n_user_dim: int
            The dimensionality of the user input vector. When using metadata, you should
            make sure to set this to the size of each of these vectors.
        n_item_dim: int
            The dimensionality of the item input vector. When using metadata, you should
            make sure to set this to the size of each of these vectors.
        n_factors: int
            The dimensionality of the latent feature space _for both users and items_
            for the GMF component of the architecture.

        Returns
        -------
        output: Model
            The 'complete' Keras Model object.

        References
        ----------
        [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569
        [2] https://github.com/hexiangnan/neural_collaborative_filtering
        """

        n_user_vocab = dataset.all_users.shape[0]
        n_item_vocab = dataset.all_items.shape[0]

        if dataset.user_meta is not None:
            n_user_vocab += dataset.user_meta.shape[1]
        if dataset.item_meta is not None:
            n_item_vocab += dataset.item_meta.shape[1]

        user_input, user_bias, user_factors = utils.get_embedding_block(
            n_user_vocab, n_user_dim, n_factors, **kwargs
        )
        item_input, item_bias, item_factors = utils.get_embedding_block(
            n_item_vocab, n_item_dim, n_factors, **kwargs
        )

        body = Multiply()([user_factors, item_factors])
        output = Dense(1, activation="sigmoid", kernel_initializer=lecun_uniform())(
            body
        )

        return Model(inputs=[user_input, item_input], outputs=output)



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

import math
import multiprocessing as mp
from functools import partial

from typing import List, Callable, Tuple, Any, Optional

from numpy import asarray, isin, unique, ndarray


def _parfn(
    fn: Callable[[List[int], List[int]], float],
    args: Tuple[Any],
    **kwargs: Optional[Any],
) -> float:
    """
    A utility function for unpacking arguments when a function is called inside a
    parallel map. You can't pickle lambdas!
    """

    return fn(*args, **kwargs)


def score(
    fn: Callable[[ndarray, ndarray, Optional[Any]], float],
    actual: List[ndarray],
    predicted: List[ndarray],
    n_cpu: int = -1,
    **kwargs: Optional[Any],
) -> ndarray:
    """
    Score a set of predicted documents (e.g. recommendations) against 'actual' relevant
    documents (e.g. held-out transaction history).

    This is currently compatible with:
    * precision_at_k
    * normalized_discounted_cumulative_gain

    For 'coverage_at_k', use that function directly.

    Parameters
    ----------
    fn: callable
        A scoring function. This should take at least two arguments: an array-like
        object containing up to 'M' relevant documents, and a second array-like object
        containing 'K' predicted relevant documents.
    actual: array-like
        An array-like object where each element contains up to 'M' elements. Note that
        this means this object may contain elements that are different lengths. This
        is expected. These are collections of relevant documents.
    predicted: array-like
        An array-like object where each element contains 'K' elements. These are each
        collections of documents that are expected to be relevant.
    n_cpu: int
        Indicate whether to utilise multiple cores. By default, n_cpu < 2 indicates the
        execution should proceed within a single process.
    kwargs: optional
        Additional keyword arguments to be passed to the provided metric 'fn'.

    Returns
    -------
    output: list
        A list of floats, where each float corresponds to the computed metric for each
        actual-predicted pair.

    """

    if len(actual) != len(predicted):
        raise ValueError(
            "Total number of results does not match the total number of "
            "expected outputs."
        )

    if n_cpu <= 2:
        output = [fn(a, p, **kwargs) for a, p in zip(actual, predicted)]
    else:
        with mp.Pool(processes=n_cpu) as pool:
            output = pool.map(partial(_parfn, fn, **kwargs), zip(actual, predicted))

    return asarray(output)


def coverage_at_k(actual: ndarray, predicted: List[ndarray], k: int = 10) -> float:
    """
    Compute the coverage at k (c@k) for a given set of documents (actual) given all
    predictions.

    This metric is useful for checking how many items from across a catalog or
    returned in the top 'k' results.

    Parameters
    ----------
    actual: list, array-like
        An array of containing all documents in a set/catalog.
    predicted: list, array-like
        An array of containing lists of predicted documents (e.g. recommendations).
    k: int
        The total number of predicted results to consider. Default 10.

    Returns
    -------
    output: float
        The c@k for the given element sets.

    """

    n = isin(unique(asarray(predicted)[:, :k]), actual).sum()
    return n / len(actual)


def precision_at_k(actual: ndarray, predicted: ndarray, k: int = 10) -> float:
    """
    Compute the precision at k (p@k) for a given set of documents.

    For example, precision at 5 corresponds to the number of relevant results among the
    top 5 expected results.

    Parameters
    ----------
    actual: list, array-like
        An array of containing relevant documents (e.g. 'actual' user activity).
    predicted: list, array-like
        An array of containing predicted documents (e.g. recommendations).
    k: int
        The total number of predicted results to consider. Default 10.

    Returns
    -------
    output: float
        The p@k for the given element sets.

    References
    ----------
    [1] https://github.com/benhamner/Metrics/blob/master/Python/ml_metrics/
    [2] https://en.wikipedia.org/wiki/Precision_and_recall#Precision

    """

    hits: float = 0.0

    for element in predicted[:k]:

        if element in actual:
            hits += 1.0

    return hits / min(len(actual), k)


def normalized_discounted_cumulative_gain(
    actual: ndarray, predicted: ndarray, k: int = 10,
) -> float:
    """
    Compute the nDCG for a set of documents assuming binary relevance scores (i.e.
    the presence of an 'actual' element in the predicted elements has relevance '1').

    Examples
    --------

    Consider a set of documents given by the indices:

    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}

    And that the following six documents are relevant to a search query (actual):

    {0, 2, 3, 5, 7, 9}

    And four query responses drawn from these documents (predicted):

    {5, 8, 3, 2}

    Then the relevance of the 4 response documents would be:

    {1, 0, 1, 1}

    The DCG would then be: SUM((1 / log(2)) + (1 / log(4)) + (1 / log(5)))
    And the IDCG: SUM((1 / log(2)) + (1 / log(3)) + (1 / log(4)) + (1 / log(5)))

    Where DCG is the Discounted Cumulative Gain, and IDCG is the Ideal DCG (best-
    possible DCG).

    Then we have the normalized metric: nDCG = DCG / IDCG

    Parameters
    ----------
    actual: list, array-like
        An array of containing relevant documents (e.g. 'actual' user activity).
    predicted: list, array-like
        An array of containing predicted documents (e.g. recommendations).
    k: int
        The total number of predicted results to consider. Default 10.

    Returns
    -------
    output: float
        The nDCG for the given element sets.

    References
    ----------
    [1] https://en.wikipedia.org/wiki/Discounted_cumulative_gain

    """

    dcg: float = 0.0
    idcg: float = 0.0
    for i, _ in enumerate(predicted[:k]):
        dcg_i = 1.0 / math.log(i + 2)

        if _ in actual:
            dcg += dcg_i

        idcg += dcg_i

    return dcg / idcg


def hit_ratio(actual: ndarray, predicted: ndarray, *_: Optional[Any]) -> float:
    """
    Compute the hit ratio (determine if a single element exists in predicted).

    Parameters
    ----------
    actual: list
        A list of at least one element, where the first element is checked to determine
        if it appears in 'predicted'.
    predicted: list
        A list (potentially unordered) of items.

    Returns
    -------
    output: float
        Indicate whether the element exists or not.

    References
    ----------
    [1] https://arxiv.org/pdf/1708.05024.pdf

    """

    if actual[0] in predicted:
        return 1.0
    else:
        return 0.0


def truncated_ndcg(actual: ndarray, predicted: ndarray, *_: Optional[Any]) -> float:
    """
    A special case of nDCG specifically to check if a single element exists in the
    predicted set, and to compute the nDCG for this item only (i.e. for a single
    element).

    This implementation reflects the implementation found in [1], as described in [2].

    Parameters
    ----------
    actual: list
        A list of at least one element, where the first element is checked to determine
        if it appears in 'predicted', and the nDCG computed for this element.
    predicted: list
        A list of ordered items.

    References
    ----------
    [1] https://github.com/hexiangnan/neural_collaborative_filtering/blob/master/MLP.py
    [2] https://arxiv.org/pdf/1708.05024.pdf

    """

    z = actual[0]
    for i, e in enumerate(predicted):
        if e == z:
            return math.log(2) / math.log(i + 2)

    return 0.0


# aliases
cak = coverage_at_k
pak = precision_at_k
ndcg = normalized_discounted_cumulative_gain



from setuptools import setup
from setuptools import find_packages

from xanthus import __version__

long_description = """
Implementations of Neural Collaborative Filtering models, based on/inspired by the work 
of He et al., built on Keras, with a few recommendation system utilities to boot. 
"""

setup(
    name="xanthus",
    version=__version__,
    description="Neural Collaborative Filtering in Python",
    long_description=long_description,
    author="Mark Douthwaite",
    author_email="mark@douthwaite.io",
    url="https://github.com/markdouthwaite/xanthus",
    license="MIT",
    install_requires=["tensorflow==2.2.0", "jupyter==1.0.0", "pandas==1.0.4", "numpy==1.18.5", "h5py==2.10.0", "scipy==1.4.1", "scikit-learn==0.23.1", "requests==2.23.0", "implicit==0.4.0"],
    extras_require={"tests": ["pytest", "pandas", "requests", "markdown", "black", "tensorflow"]},
    classifiers=[
        # 'Development Status :: 5 - Production/Stable',
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(),
)



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from itertools import islice
from typing import Optional, Tuple, Iterable, Any

from tensorflow.keras.layers import (
    Layer,
    Input,
    Flatten,
    Embedding,
)
from tensorflow.keras.regularizers import l2
from tensorflow.keras.initializers import RandomNormal, Initializer


def get_embedding_block(
    n_vocab: int,
    n_dim: int,
    n_factors: int,
    reg: float = 0.0,
    initializer: Initializer = RandomNormal(stddev=0.01),
    inputs: Optional[Input] = None,
) -> Tuple[Layer, Embedding, Embedding]:
    """
    Utility function for building Embedding blocks for neural recommenders.

    Parameters
    ----------
    n_vocab: int
        The size of the 'vocabulary' of the Embedding.
    n_dim: int
        The length of the input vectors.
    n_factors: int
        The total number of latent features in the recommender model (output).
    reg: float
        The L2 regularization to apply to the embeddings.
    initializer: Initializer
        The weight initializer for the embeddings.
    inputs: Input, optional
        Optionally provide an 'input' layer for the embedding block.

    Returns
    -------
    output: tuple
        A tuple, where the first element corresponds to the Embeddings' input layer,
        the second a bias term, and the third the embedding layer itself.

    """

    inputs = inputs if inputs is not None else Input(shape=(n_dim,))
    embedding = Embedding(
        input_dim=n_vocab,
        output_dim=n_factors,
        input_length=n_dim,
        embeddings_initializer=initializer,
        embeddings_regularizer=l2(reg),
    )(inputs)

    bias_embedding = Embedding(input_dim=n_vocab, output_dim=1)(inputs)
    factors = Flatten()(embedding)

    return inputs, Flatten()(bias_embedding), factors


def batched(i: Iterable[Any], n: int) -> Iterable[Any]:
    """Batch an iterable 'i' into batches of 'n'."""

    g = (_ for _ in i)
    c = list(islice(g, n))
    while c:
        yield c
        c = list(islice(g, n))



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from abc import ABC, abstractmethod
from typing import List, Optional, Any, Dict, Callable, Tuple

import numpy as np
from numpy import ndarray

from tensorflow.keras import Model
from tensorflow.keras.optimizers import Adam, Optimizer
from tensorflow.keras.losses import BinaryCrossentropy, Loss

from sklearn.model_selection import train_test_split

from xanthus.datasets import Dataset


Metric = Callable[[List[int], List[int], Optional[Any]], float]


class RecommenderModel(ABC):
    """A simple recommender model interface. Inspired by the scikit-learn API."""

    @abstractmethod
    def fit(self, dataset: Dataset) -> "RecommenderModel":
        """Fit the model on the provided dataset."""

    @abstractmethod
    def predict(
        self,
        dataset: Dataset,
        users: Optional[List[int]] = None,
        items: Optional[List[int]] = None,
        n: int = 3,
        excluded: Optional[List[int]] = None,
    ) -> List[List[int]]:
        """
        Generate 'n' predictions for the given dataset (and optionally, provided users
        and items).

        Parameters
        ----------
        dataset: Dataset
            The dataset to be used for generating recommendations.
        users: list, optional
            An optional list of users for whom you wish to generate recommendations.
        items: list, optional
            An optional list of items for which you wish to generate recommendations.
            This should map to individual users (either those in 'users' or in
            dataset.all_users).
        n: int
            The total number of recommendations for each user.
        excluded: list, optional
            An optional list of items to exclude from recommendations.

        Returns
        -------
        output: list
            A list, where each element corresponds to a list of recommendations. If a
            list of 'users' was provided, this will be ordered by this list. If 'users'
            are not provided, it will be ordered by 'dataset.all_users'.

        """


class NeuralRecommenderModel(RecommenderModel):
    """
    RecommenderModels with a Neural flavour, designed to support negative sampling as
    described in [1], and to work with Xanthus Datasets. Built with Keras.

    Parameters
    ----------
    loss: Loss
        A Keras Loss function of your choice. BinaryCrossEntropy is used by default [1].
    optimizer: Optimizer
        A Keras Optimizer of your choice. Adam is used by default [1].
    negative_samples: int
        Specify the number of negative samples (per positive sample) you wish to
        generate during training. Note that negative sampling can dramatically slow
        down your training loop, but can also boost performance in the end.
    fit_params: dict, optional
        Optional fit parameters to be passed to the Keras `fit` call.
    kwargs: any, optional
        Additional keyword arguments.

    Notes
    -----
    * Supports negative sampling. This is currently v. expensive.
    * Subclassing of a Keras Model was considered, but rejected as it limits your
      ability to save a model in a format other than `pickle`.

    References
    ----------
    [1] He et al. https://dl.acm.org/doi/10.1145/3038912.3052569

    See Also
    --------
    tensorflow.ketas.Model.fit


    """

    def __init__(
        self,
        loss: Loss = BinaryCrossentropy(),
        optimizer: Optimizer = Adam(lr=1e-3),
        negative_samples: int = 0,
        n_meta: int = 0,
        fit_params: Optional[Dict[str, Any]] = None,
        **kwargs: Optional[Any],
    ):
        """
        Initialize a NeuralRecommender
        """

        self.model = None
        self._loss = loss
        self._optimizer = optimizer
        self._negative_samples = negative_samples
        self._config = kwargs
        self._n_meta = n_meta
        self._fit_params = fit_params

    def fit(
        self, dataset: Dataset, **kwargs: Optional[Any],
    ) -> "NeuralRecommenderModel":
        """
        Fit the model to a provided Dataset.

        Parameters
        ----------
        dataset: Dataset
            An input dataset.

        Returns
        -------
        output: NeuralRecommenderModel
            Returns itself.

        See Also
        --------
        xanthus.datasets.Dataset

        """

        n_dim = self._n_meta + 1

        if self.model is None:
            self.model = self._build_model(
                dataset, n_user_dim=n_dim, n_item_dim=n_dim, **self._config
            )
            self.model.compile(optimizer=self._optimizer, loss=self._loss)

        epochs, fit_params = self._unpack_fit_params()

        for i in range(epochs):
            user_x, item_x, y = dataset.to_components(
                negative_samples=self._negative_samples, output_dim=self._n_meta + 1
            )

            tux, vux, tix, vix, ty, vy = train_test_split(
                user_x, item_x, y, test_size=0.1
            )
            print(f"Epoch {i+1}/{epochs}")
            self.model.fit(
                [tux, tix],
                ty,
                epochs=1,
                **fit_params,
                validation_data=([vux, vix], vy),
                **kwargs,
            )

        return self

    def predict(
        self,
        dataset: Optional[Dataset] = None,
        users: Optional[ndarray] = None,
        items: Optional[ndarray] = None,
        n: int = 3,
        excluded: Optional[ndarray] = None,
    ) -> List[List[int]]:
        """
        Generate predictions (recommendations) from the model. For each provided user,
        the output will be a set of items ranked in order of predicted preference.

        Parameters
        ----------
        dataset: Dataset, optional
            The dataset for which you wish to generate recommendations. If this is
            provided, this object's `all_items` and `users` will be used for the
            purpose of generating recommendations.
        users: list, optional
            An optional array of users for whom you wish to generate recommendations.
        items: list, optional
            An optional array of items you wish to be used in recommendations. This
            may be a subset of items for the purposes of ranking specific subsets of
            items, for example. Maybe you want to see which Star Wars movies are the
            most loved according to the model, for example.
        n: int
            The number of recommendations to be generated per user.
        excluded: list, optional
            An optional array of items to exclude from recommendations.
        Returns
        -------
        output: list
            A list, where each element corresponds to a list of recommendations. If a
            list of 'users' was provided, this will be ordered by this list. If 'users'
            are not provided, it will be ordered by 'dataset.all_users'.

        # Todo:
        # * This needs an overhaul. Some of that will involve re-working the
        #   dataset generators, but this feels v. clunky & extraordinarily slow.

        """

        recommended = []

        users = users if users is not None else dataset.users

        if items is None:
            all_items = dataset.all_items
            items = (all_items for _ in users)

        if self._n_meta > 0:
            items = (dataset.iter_item(e, n_dim=self._n_meta + 1) for e in items)
            users = dataset.iter_user(users, n_dim=self._n_meta + 1)

        for (user, target_items) in zip(users, items):
            # this _list_ unpacks the metadata.
            target_items = np.asarray(list(target_items))

            if len(target_items.shape) == 1:
                target_items = target_items.reshape(-1, 1)

            x = [np.tile(user, (len(target_items), 1)), target_items]

            h = self.model(x).numpy().flatten()
            ranked = self._rank(h, n, encodings=target_items[:, 0], excluded=excluded)

            recommended.append(ranked[:n])

        return recommended

    @abstractmethod
    def _build_model(self, dataset: Dataset, **kwargs: Optional[Any]) -> Model:
        """Build a Keras model."""

    def _unpack_fit_params(self, epochs: int = 1) -> Tuple[int, Dict[str, Any]]:
        """Unpack fit parameters, extracting `epochs` for use in training loop."""

        epochs = self._fit_params.get("epochs", epochs)
        fit_params = {k: v for k, v in self._fit_params.items() if k != "epochs"}
        return epochs, fit_params

    @staticmethod
    def _rank(
        w: ndarray,
        n: int,
        encodings: Optional[ndarray] = None,
        excluded: Optional[ndarray] = None,
    ) -> ndarray:
        """
        A utility for simply ranking predictions.

        Parameters
        ----------
        w: ndarray
            The input weights/scores to rank.
        n: int
            The total number of recommendations you require.
        encodings: ndarray, optional
            Optional encodings that should map to the elements in the `w` array.
        excluded: ndarray, optional
            Optional elements that should be excluded from the output rankings.

        Returns
        -------
        output: ndarray
            An array at most `n` elements in length.

        """

        ranked = w.argsort()[::-1]

        if encodings is not None:
            ranked = encodings[ranked]

        if excluded is not None:
            ranked = ranked[~np.isin(ranked, excluded)]

        return ranked[:n]

    def __repr__(self):
        return f"{type(self).__name__}()"



"""
The MIT License

Copyright (c) 2018-2020 Mark Douthwaite
"""

from xanthus.models.neural import (
    GeneralizedMatrixFactorizationModel,
    NeuralMatrixFactorizationModel,
    MultiLayerPerceptronModel,
)
from xanthus.models.baseline import MatrixFactorizationModel, PopRankModel
